// A crude utility to convert SMS messages saved as CSV files from Nokia PC 
// Suite (probably Ovi Suite as well) into XML files readable by 'SMS  
// Backup and Restore' app on Android.
//
// Matt Day, July 2010
//
// You are free to use this source code as you see fit, but unless otherwise 
// stated, this software is provided "as-is," without any express or implied 
// warranty. In no event shall the author be held liable for any damages 
// arising from the use of the software.

#include <cstdlib>
#include <string>
#include <iostream>
#include <fstream>
#include <vector>
#include <ctime>
#include <sstream>

const char *c_appName = "nsms2xml";

using namespace std;

//-----------------------------------------------------------------------------
// An SMS message
struct SMS {
	string type;
	enum {
		Received,
		Sent,
	} direction;
	string number;
	string subject;
	string timestamp;
	string dummy;
	string message;
};
//-----------------------------------------------------------------------------
// Retrieve next data field from line, incrementing offset.
// The messageBody flag indicates the field should be treated as the SMS 
// message content
string nextField(const string &line, size_t &offset, bool messageBody = false) {

	string field;
	size_t beginOffset = offset;
	size_t endOffset = line.find(',',beginOffset);
	offset = endOffset + 1;
	if (messageBody) endOffset = line.size();
	field = line.substr(beginOffset, endOffset - beginOffset);
	if (field[0] == '"' && field[field.size() - 1] == '"')
		field = field.substr(1, field.size() - 2);
	if (messageBody) {
		// Encode special characters where necessary
		size_t specialCharOffset = 0;
		while (specialCharOffset != string::npos) {
			beginOffset = specialCharOffset;
			specialCharOffset = field.find('&',beginOffset);
			if (specialCharOffset == string::npos)
				specialCharOffset = field.find('<',beginOffset);
			if (specialCharOffset == string::npos)
				specialCharOffset = field.find('>',beginOffset);
			if (specialCharOffset == string::npos)
				specialCharOffset = field.find('\\',beginOffset);
			if (specialCharOffset == string::npos)
				specialCharOffset = field.find('"',beginOffset);
			if (specialCharOffset != string::npos) {
				char c = field[specialCharOffset];
				field[specialCharOffset] = '&';
				switch(c) {
					case '&':
						field.insert(specialCharOffset + 1, "amp;");
						specialCharOffset += 4;
						break;
					case '<':
						field.insert(specialCharOffset + 1, "lt;");
						specialCharOffset += 3;
						break;
					case '>':
						field.insert(specialCharOffset + 1, "gt;");
						specialCharOffset += 3;
						break;
					case '"':
						// This is a bit of a hack, since it seems quote marks come out as double in the Nokia csv
						field.insert(specialCharOffset + 1, "quot");
						specialCharOffset += 5;
						field[specialCharOffset]= ';';
						break;
				}
				specialCharOffset ++;
			}
		}
	}
	return field;

}
//-----------------------------------------------------------------------------
// Parse one line of csv file, storing in SMS structure
bool parseSMS(SMS &sms, const string &line) {

	size_t offset = 0;
	string direction;
	sms.type = nextField(line, offset);
	if (sms.type.compare("sms")) return false;
	direction = nextField(line, offset);
	if (direction.compare("deliver") == 0)
		sms.direction = SMS::Received;
	else if (direction.compare("submit") == 0) 
		sms.direction = SMS::Sent;
	if (sms.direction == SMS::Received) {
		sms.number = nextField(line, offset);
		direction = nextField(line, offset);
	}
	else {
		direction = nextField(line, offset);
		sms.number = nextField(line, offset);
	}
	sms.subject = nextField(line, offset);
	sms.timestamp = nextField(line, offset);
	sms.dummy = nextField(line, offset);
	sms.message = nextField(line, offset, true);
	return true;
	
}
//-----------------------------------------------------------------------------
// Convert time from Nokia scsv to Unix epoch time
string toEpochTime(const string &time) {

	struct tm calendarTime;
	calendarTime.tm_year = atoi(time.substr(0,4).c_str()) - 1900;
	calendarTime.tm_mon = atoi(time.substr(5,2).c_str()) - 1;
	calendarTime.tm_mday = atoi(time.substr(8,2).c_str());
	calendarTime.tm_hour = atoi(time.substr(11,2).c_str());
	calendarTime.tm_min = atoi(time.substr(14,2).c_str());
	calendarTime.tm_sec = 0;
	calendarTime.tm_isdst = 0;
	time_t epochTime = mktime(&calendarTime);
	stringstream s;
	s << epochTime;
	return s.str();

}
//-----------------------------------------------------------------------------
// Write vector of SMS messages to output
void writeXML(ofstream &output, const vector<SMS> &vSMS) {

	output << "<?xml version='1.0' encoding='UTF-8' standalone='yes' ?>\n";
	output << "<smses count=\"" << vSMS.size() << "\">\n";
	for (size_t i = 0; i < vSMS.size(); ++ i) {
		output << "<sms protocol=\"0\" address=\"" << vSMS[i].number;
		output << "\" date=\"";
		output << toEpochTime(vSMS[i].timestamp) << "000\" type=\"";
		if (vSMS[i].direction == SMS::Sent) output << "2\"";
		else output << "1\"";
		output << " subject=\"null\" body=\"";
		output << vSMS[i].message << "\" toa=\"0\" sc_toa=\"0\" service_center=\"null\" read=\"1\" status=\"-1\" />\n";

	}
	output << "</smses>";

}
//-----------------------------------------------------------------------------
// Entry point
int main(const int argc, const char **argv) {

	if (argc != 3) {
		cout << c_appName << " is a command-line utility for converting SMS "
			"message data from the \nCSV format used by Nokia PC Suite to the "
			"XML format used by the Android \n'SMS Backup and Restore' app.\n\n";
		cout << "Usage: " << c_appName << " inputFile outputFile\n";
		return -1;
	}

	ifstream input(argv[1]);
	ofstream output(argv[2]);

	if (!input.is_open()) {
		cout << "Could not open " << argv[1] << endl;
		return -1;
	}
	if (!output.is_open()) {
		cout << "Could not open " << argv[2] << endl;
		return -1;
	}

	vector<SMS> vSMS;
	string line;

	// Read messages
	while (!input.eof()) {	
		SMS sms;
		getline(input, line);
		if (parseSMS(sms, line)) vSMS.push_back(sms);
	}
	// Write messages
	writeXML(output, vSMS);

	cout << "Converted " << vSMS.size() << " messages\n";

	input.close();
	output.close();
	return 0;

}
//-----------------------------------------------------------------------------
